/*
 * Client code generation.
 *
 * Copyright (C) 2019, 2020, 2021 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <stdlib.h>
#include "client.h"
#include "config.h"
#include "common.h"
#include "declaration.h"
#include "message.h"
#include "templates.h"



/* Generate operation signature details for client use. */

void write_client_signature(struct signature *sig, enum signature_role role,
                            FILE *fp, struct interface *iface)
{
  /* Client interface signatures are oblivious to completions. */

  write_interface_signature(sig, role, CLIENT_ROLE, GENERAL_FUNCTION_ROLE, fp,
                            iface);
}

/* Generate function source code for each operation signature. */

void write_client_function(struct signature *sig, FILE *fp, struct interface *iface)
{
  write_client_signature(sig, DEFINITION_ROLE, fp, iface);
  write_client_function_body(sig->parameters, fp, iface, sig);
  fputs(END_FUNCTION, fp);
}

/* Generate a function body corresponding to an operation for client use. */

void write_client_function_body(struct parameter *param, FILE *fp,
                                struct interface *iface, struct signature *sig)
{
  int input_words, input_items, output_words, output_items;

  /* With a protocol attribute, the opcode is the attribute value. Otherwise,
     it is the operation code. */

  char *opname = get_operation_name(iface, sig),
       *protocol = get_protocol(iface->attributes),
       *opcode = get_opcode_identifier(protocol, opname);

  /* Generate the prologue. */

  fputs(client_function_body_prologue, fp);

  /* Obtain a suitable endpoint reference. */

  if (conf.language == C_LANGUAGE)
    fprintf(fp, client_function_body_endpoint_conversion_c, L4_CAP_TYPE);

  /* Count the number of input words, items and output words. */

  count_parameters(param, &input_words, &input_items, &output_words, &output_items);

  /* Adjust the input words where a protocol is used as the opcode. */

  if (protocol != NULL)
    input_words++;

  /* Generate types parameterised with the qualified operation name. */

  write_accessor_declaration(input_words, output_words, opname, fp);

  /* Initialise the message. */

  fprintf(fp, "\n  ipc_message_new(&msg);\n");

  /* Generate expected output item requirements. */

  if (output_items)
    fprintf(fp, "  err = ipc_message_expect(&msg, %d);\n"
                "  if (err)\n"
                "    return err;\n", output_items);

  /* Populate input parameters in the message. Dereference function parameters
     acting as "inout" parameters. */

  if (input_words)
  {
    /* Reserve space for the words before any items, obtaining a pointer to the
       words. */

    write_accessor_initialisation(IN_PARAMETER, CLIENT_ROLE, GENERAL_FUNCTION_ROLE,
                                  opname, fp);

    if (protocol != NULL)
      fprintf(fp, "  in_words->_op = opcode_%s;\n", opname);

    write_message_access(param, CLIENT_ROLE, GENERAL_FUNCTION_ROLE, IN_PARAMETER,
                         WORD_CLASS, fp);
  }

  /* Do the same for items. */

  if (input_items)
  {
    fputs("\n", fp);
    write_message_access(param, CLIENT_ROLE, GENERAL_FUNCTION_ROLE, IN_PARAMETER,
                         ITEM_CLASS, fp);
  }

  /* Send the request. */

  fprintf(fp, client_function_body_call, opcode, "_endp");

  /* Retrieve output parameters from the message. */

  if (output_words)
  {
    write_accessor_initialisation(OUT_PARAMETER, CLIENT_ROLE, GENERAL_FUNCTION_ROLE,
                                  opname, fp);
    write_message_access(param, CLIENT_ROLE, GENERAL_FUNCTION_ROLE, OUT_PARAMETER,
                         WORD_CLASS, fp);
  }

  /* Do the same for items. */

  if (output_items)
  {
    fputs("\n", fp);
    write_message_access(param, CLIENT_ROLE, GENERAL_FUNCTION_ROLE, OUT_PARAMETER,
                         ITEM_CLASS, fp);
  }

  /* Generate the epilogue. */

  fputs(client_function_body_epilogue, fp);

  /* Free allocated strings. */

  free(opname);

  if (protocol == NULL)
    free(opcode);
}

/* Populate an interface structure with client functions. */

void write_client_interface(struct interface *iface, FILE *fp)
{
  struct signature *sig;
  char *opname;

  if (iface->signatures == NULL)
    fprintf(fp, client_interface_prologue_empty_c, iface->name, iface->name);
  else
  {
    fprintf(fp, client_interface_prologue_c, iface->name, iface->name);

    for (sig = iface->signatures; sig != NULL; sig = sig->tail)
    {
      opname = get_operation_name(iface, sig);

      fprintf(fp, client_interface_member_c, sig->operation, opname);

      free(opname);
    }

    fputs(client_interface_epilogue_c, fp);
  }
}
